@@ -2,7 +2,7 @@
  * The University of Illinois/NCSA
  * Open Source License (NCSA)
  *
- * Copyright (c) 2017, Advanced Micro Devices, Inc.
+ * Copyright (c) 2017-2023, Advanced Micro Devices, Inc.
  * All rights reserved.
  *
  * Developed by:
@@ -57,6 +57,7 @@
 #include <cerrno>
 #include <unordered_map>
 #include <iostream>
+#include <sstream>
 
 #include "rocm_smi/rocm_smi.h"
 #include "rocm_smi/rocm_smi_device.h"
@@ -64,6 +65,9 @@
 #include "rocm_smi/rocm_smi_exception.h"
 #include "rocm_smi/rocm_smi_utils.h"
 #include "rocm_smi/rocm_smi_kfd.h"
+#include "rocm_smi/rocm_smi_logger.h"
+
+using namespace ROCmLogging;
 
 static const char *kPathDRMRoot = "/sys/class/drm";
 static const char *kPathHWMonRoot = "/sys/class/hwmon";
@@ -73,6 +77,81 @@ static const char *kDeviceNamePrefix = "card";
 
 static const char *kAMDMonitorTypes[] = {"radeon", "amdgpu", ""};
 
+static const std::string amdSMI = "amd::smi::";
+const std::map<amd::smi::DevInfoTypes, std::string>
+amd::smi::RocmSMI::devInfoTypesStrings = {
+  {amd::smi::kDevPerfLevel, amdSMI + "kDevPerfLevel"},
+  {amd::smi::kDevOverDriveLevel, amdSMI + "kDevOverDriveLevel"},
+  {amd::smi::kDevMemOverDriveLevel, amdSMI + "kDevMemOverDriveLevel"},
+  {amd::smi::kDevDevID, amdSMI + "kDevDevID"},
+  {amd::smi::kDevDevProdName, amdSMI + "kDevDevProdName"},
+  {amd::smi::kDevDevProdNum, amdSMI + "kDevDevProdNum"},
+  {amd::smi::kDevVendorID, amdSMI + "kDevVendorID"},
+  {amd::smi::kDevSubSysDevID, amdSMI + "kDevSubSysDevID"},
+  {amd::smi::kDevSubSysVendorID, amdSMI + "kDevSubSysVendorID"},
+  {amd::smi::kDevGPUMClk, amdSMI + "kDevGPUMClk"},
+  {amd::smi::kDevGPUSClk, amdSMI + "kDevGPUSClk"},
+  {amd::smi::kDevDCEFClk, amdSMI + "kDevDCEFClk"},
+  {amd::smi::kDevFClk, amdSMI + "kDevFClk"},
+  {amd::smi::kDevSOCClk, amdSMI + "kDevSOCClk"},
+  {amd::smi::kDevPCIEClk, amdSMI + "kDevPCIEClk"},
+  {amd::smi::kDevPowerProfileMode, amdSMI + "kDevPowerProfileMode"},
+  {amd::smi::kDevUsage, amdSMI + "kDevUsage"},
+  {amd::smi::kDevPowerODVoltage, amdSMI + "kDevPowerODVoltage"},
+  {amd::smi::kDevVBiosVer, amdSMI + "kDevVBiosVer"},
+  {amd::smi::kDevPCIEThruPut, amdSMI + "kDevPCIEThruPut"},
+  {amd::smi::kDevErrCntSDMA, amdSMI + "kDevErrCntSDMA"},
+  {amd::smi::kDevErrCntUMC, amdSMI + "kDevErrCntUMC"},
+  {amd::smi::kDevErrCntGFX, amdSMI + "kDevErrCntGFX"},
+  {amd::smi::kDevErrCntMMHUB, amdSMI + "kDevErrCntMMHUB"},
+  {amd::smi::kDevErrCntPCIEBIF, amdSMI + "kDevErrCntPCIEBIF"},
+  {amd::smi::kDevErrCntHDP, amdSMI + "kDevErrCntHDP"},
+  {amd::smi::kDevErrCntXGMIWAFL, amdSMI + "kDevErrCntXGMIWAFL"},
+  {amd::smi::kDevErrCntFeatures, amdSMI + "kDevErrCntFeatures"},
+  {amd::smi::kDevMemTotGTT, amdSMI + "kDevMemTotGTT"},
+  {amd::smi::kDevMemTotVisVRAM, amdSMI + "kDevMemTotVisVRAM"},
+  {amd::smi::kDevMemTotVRAM, amdSMI + "kDevMemTotVRAM"},
+  {amd::smi::kDevMemUsedGTT, amdSMI + "kDevMemUsedGTT"},
+  {amd::smi::kDevMemUsedVisVRAM, amdSMI + "kDevMemUsedVisVRAM"},
+  {amd::smi::kDevMemUsedVRAM, amdSMI + "kDevMemUsedVRAM"},
+  {amd::smi::kDevVramVendor, amdSMI + "kDevVramVendor"},
+  {amd::smi::kDevPCIEReplayCount, amdSMI + "kDevPCIEReplayCount"},
+  {amd::smi::kDevUniqueId, amdSMI + "kDevUniqueId"},
+  {amd::smi::kDevDFCountersAvailable, amdSMI + "kDevDFCountersAvailable"},
+  {amd::smi::kDevMemBusyPercent, amdSMI + "kDevMemBusyPercent"},
+  {amd::smi::kDevXGMIError, amdSMI + "kDevXGMIError"},
+  {amd::smi::kDevFwVersionAsd, amdSMI + "kDevFwVersionAsd"},
+  {amd::smi::kDevFwVersionCe, amdSMI + "kDevFwVersionCe"},
+  {amd::smi::kDevFwVersionDmcu, amdSMI + "kDevFwVersionDmcu"},
+  {amd::smi::kDevFwVersionMc, amdSMI + "kDevFwVersionMc"},
+  {amd::smi::kDevFwVersionMe, amdSMI + "kDevFwVersionMe"},
+  {amd::smi::kDevFwVersionMec, amdSMI + "kDevFwVersionMec"},
+  {amd::smi::kDevFwVersionMec2, amdSMI + "kDevFwVersionMec2"},
+  {amd::smi::kDevFwVersionPfp, amdSMI + "kDevFwVersionPfp"},
+  {amd::smi::kDevFwVersionRlc, amdSMI + "kDevFwVersionRlc"},
+  {amd::smi::kDevFwVersionRlcSrlc, amdSMI + "kDevFwVersionRlcSrlc"},
+  {amd::smi::kDevFwVersionRlcSrlg, amdSMI + "kDevFwVersionRlcSrlg"},
+  {amd::smi::kDevFwVersionRlcSrls, amdSMI + "kDevFwVersionRlcSrls"},
+  {amd::smi::kDevFwVersionSdma, amdSMI + "kDevFwVersionSdma"},
+  {amd::smi::kDevFwVersionSdma2, amdSMI + "kDevFwVersionSdma2"},
+  {amd::smi::kDevFwVersionSmc, amdSMI + "kDevFwVersionSmc"},
+  {amd::smi::kDevFwVersionSos, amdSMI + "kDevFwVersionSos"},
+  {amd::smi::kDevFwVersionTaRas, amdSMI + "kDevFwVersionTaRas"},
+  {amd::smi::kDevFwVersionTaXgmi, amdSMI + "kDevFwVersionTaXgmi"},
+  {amd::smi::kDevFwVersionUvd, amdSMI + "kDevFwVersionUvd"},
+  {amd::smi::kDevFwVersionVce, amdSMI + "kDevFwVersionVce"},
+  {amd::smi::kDevFwVersionVcn, amdSMI + "kDevFwVersionVcn"},
+  {amd::smi::kDevSerialNumber, amdSMI + "kDevSerialNumber"},
+  {amd::smi::kDevMemPageBad, amdSMI + "kDevMemPageBad"},
+  {amd::smi::kDevNumaNode, amdSMI + "kDevNumaNode"},
+  {amd::smi::kDevGpuMetrics, amdSMI + "kDevGpuMetrics"},
+  {amd::smi::kDevGpuReset, amdSMI + "kDevGpuReset"},
+  {amd::smi::kDevAvailableComputePartition, amdSMI +
+      "kDevAvailableComputePartition"},
+  {amd::smi::kDevComputePartition, amdSMI + "kDevComputePartition"},
+  {amd::smi::kDevMemoryPartition, amdSMI + "kDevMemoryPartition"}
+};
+
 namespace amd {
 namespace smi {
 
@@ -179,18 +258,21 @@ static bool bdfid_from_path(const std::string in_name, uint64_t *bdfid) {
   return true;
 }
 
+// 0 = successful bdfid found
+// 1 = not a good bdfid found
 static uint32_t ConstructBDFID(std::string path, uint64_t *bdfid) {
   assert(bdfid != nullptr);
-  char tpath[256] = {'\0'};
+  const unsigned int MAX_BDF_LENGTH = 512;
+  char tpath[MAX_BDF_LENGTH] = {'\0'};
   ssize_t ret;
-  memset(tpath,0,256);
+  memset(tpath,0,MAX_BDF_LENGTH);
 
-  ret = readlink(path.c_str(), tpath, 256);
+  ret = readlink(path.c_str(), tpath, MAX_BDF_LENGTH);
 
   assert(ret > 0);
-  assert(ret < 256);
+  assert(ret < MAX_BDF_LENGTH);
 
-  if (ret <= 0 || ret >= 256) {
+  if (ret <= 0 || ret >= MAX_BDF_LENGTH) {
     return 1;
   }
 
@@ -221,6 +303,16 @@ RocmSMI::Initialize(uint64_t flags) {
   uint32_t ret;
   int i_ret;
 
+  LOG_ALWAYS("=============== ROCM SMI initialize ================");
+  Logger::getInstance()->enableAllLogLevels();
+  // Leaving below to allow developers to check current log settings
+  // std::string logSettings = Logger::getInstance()->getLogSettings();
+  // std::cout << "Current log settings:\n" << logSettings << std::endl;
+
+  if (Logger::getInstance()->isLoggerEnabled()) {
+    logSystemDetails();
+  }
+
   assert(ref_count_ == 1);
   if (ref_count_ != 1) {
     throw amd::smi::rsmi_exception(RSMI_INITIALIZATION_ERROR,
@@ -232,8 +324,8 @@ RocmSMI::Initialize(uint64_t flags) {
   euid_ = geteuid();
 
   GetEnvVariables();
-
-  while (env_vars_.debug_inf_loop) {}
+  // To help debug env variable issues
+  // printEnvVarInfo();
 
   while (std::string(kAMDMonitorTypes[i]) != "") {
       amd_monitor_types_.insert(kAMDMonitorTypes[i]);
@@ -253,8 +345,9 @@ RocmSMI::Initialize(uint64_t flags) {
     if (ConstructBDFID(devices_[i]->path(), &bdfid) != 0) {
       std::cerr << "Failed to construct BDFID." << std::endl;
       ret = 1;
+    } else {
+      devices_[i]->set_bdfid(bdfid);
     }
-    devices_[i]->set_bdfid(bdfid);
   }
   if (ret != 0) {
     throw amd::smi::rsmi_exception(RSMI_INITIALIZATION_ERROR,
@@ -296,6 +389,7 @@ RocmSMI::Initialize(uint64_t flags) {
   // 1. construct kfd_node_map_ with gpu_id as key and *Device as value
   // 2. for each kfd node, write the corresponding dv_ind
   // 3. for each amdgpu device, write the corresponding gpu_id
+  // 4. for each amdgpu device, attempt to store it's boot partition
   for (uint32_t dv_ind = 0; dv_ind < devices_.size(); ++dv_ind) {
     dev = devices_[dv_ind];
     uint64_t bdfid = dev->bdfid();
@@ -310,7 +404,12 @@ RocmSMI::Initialize(uint64_t flags) {
     uint64_t gpu_id = tmp_map[bdfid]->gpu_id();
     dev->set_kfd_gpu_id(gpu_id);
     kfd_node_map_[gpu_id] = tmp_map[bdfid];
+
+    // store each device boot partition state, if file doesn't exist
+    dev->storeDevicePartitions(dv_ind);
   }
+  // Leaving below to help debug temp file issues
+  // displayAppTmpFilesContent();
 }
 
 void
@@ -356,23 +455,61 @@ static uint32_t GetEnvVarUInteger(const char *ev_str) {
   return 0;
 }
 
+// provides a way to get env variable detail in both debug & release
+// helps enable full logging
+static bool getRSMIEnvVar_LoggingEnabled(const char *ev_str) {
+  bool isLoggingEnabled = false;
+  ev_str = getenv(ev_str);
+
+  if (ev_str != nullptr) {
+    isLoggingEnabled = true;
+  }
+  return isLoggingEnabled;
+}
+
+static std::unordered_set<uint32_t> GetEnvVarUIntegerSets(const char *ev_str) {
+  std::unordered_set<uint32_t> returnSet;
+#ifndef DEBUG
+  (void)ev_str;
+#else
+  ev_str = getenv(ev_str);
+  if(ev_str == nullptr) { return returnSet; }
+  std::string stringEnv = ev_str;
+
+  if (stringEnv.empty() == false) {
+    // parse out values by commas
+    std::string parsedVal;
+    std::istringstream ev_str_ss(stringEnv);
+
+    while (std::getline(ev_str_ss, parsedVal, ',')) {
+      int parsedInt = std::stoi(parsedVal);
+      assert(parsedInt >= 0);
+      uint32_t parsedUInt = static_cast<uint32_t>(parsedInt);
+      returnSet.insert(parsedUInt);
+    }
+  }
+#endif
+  return returnSet;
+}
+
 // Get and store env. variables in this method
 void RocmSMI::GetEnvVariables(void) {
+  env_vars_.logging_on = getRSMIEnvVar_LoggingEnabled("RSMI_LOGGING");
 #ifndef DEBUG
   (void)GetEnvVarUInteger(nullptr);  // This is to quiet release build warning.
   env_vars_.debug_output_bitfield = 0;
   env_vars_.path_DRM_root_override = nullptr;
   env_vars_.path_HWMon_root_override = nullptr;
   env_vars_.path_power_root_override = nullptr;
-  env_vars_.enum_override = 0;
   env_vars_.debug_inf_loop = 0;
+  env_vars_.enum_overrides.clear();
 #else
   env_vars_.debug_output_bitfield = GetEnvVarUInteger("RSMI_DEBUG_BITFIELD");
   env_vars_.path_DRM_root_override   = getenv("RSMI_DEBUG_DRM_ROOT_OVERRIDE");
   env_vars_.path_HWMon_root_override = getenv("RSMI_DEBUG_HWMON_ROOT_OVERRIDE");
   env_vars_.path_power_root_override = getenv("RSMI_DEBUG_PP_ROOT_OVERRIDE");
-  env_vars_.enum_override = GetEnvVarUInteger("RSMI_DEBUG_ENUM_OVERRIDE");
   env_vars_.debug_inf_loop = GetEnvVarUInteger("RSMI_DEBUG_INFINITE_LOOP");
+  env_vars_.enum_overrides = GetEnvVarUIntegerSets("RSMI_DEBUG_ENUM_OVERRIDE");
 #endif
 }
 
@@ -380,6 +517,53 @@ const RocmSMI_env_vars& RocmSMI::getEnv(void) {
   return env_vars_;
 }
 
+bool RocmSMI::isLoggingOn(void) {
+  GetEnvVariables();
+  return this->env_vars_.logging_on;
+}
+
+void RocmSMI::printEnvVarInfo(void) {
+  std::cout << __PRETTY_FUNCTION__ << " | env_vars_.debug_output_bitfield = "
+            << ((env_vars_.debug_output_bitfield == 0) ? "<undefined>"
+                : std::to_string(env_vars_.debug_output_bitfield))
+            << std::endl;
+  std::cout << __PRETTY_FUNCTION__ << " | env_vars_.path_DRM_root_override = "
+            << ((env_vars_.path_DRM_root_override == nullptr)
+                ? "<undefined>" : env_vars_.path_DRM_root_override)
+            << std::endl;
+  std::cout << __PRETTY_FUNCTION__ << " | env_vars_.path_HWMon_root_override = "
+            << ((env_vars_.path_HWMon_root_override == nullptr)
+                ? "<undefined>" : env_vars_.path_HWMon_root_override)
+            << std::endl;
+  std::cout << __PRETTY_FUNCTION__ << " | env_vars_.path_power_root_override = "
+            << ((env_vars_.path_power_root_override == nullptr)
+                ? "<undefined>" : env_vars_.path_power_root_override)
+            << std::endl;
+  std::cout << __PRETTY_FUNCTION__ << " | env_vars_.debug_inf_loop = "
+            << ((env_vars_.debug_inf_loop == 0) ? "<undefined>"
+                : std::to_string(env_vars_.debug_inf_loop))
+            << std::endl;
+  bool isLoggingOn = (env_vars_.logging_on) ? true : false;
+  std::cout << __PRETTY_FUNCTION__ << " | env_vars_.logging_on = "
+            << (isLoggingOn ? "true" : "false") << std::endl;
+  std::cout << __PRETTY_FUNCTION__ << " | env_vars_.enum_overrides = {";
+  if (env_vars_.enum_overrides.empty()) {
+    std::cout << "}" << std::endl;
+    return;
+  }
+  for (auto it=env_vars_.enum_overrides.begin();
+       it != env_vars_.enum_overrides.end(); ++it) {
+    DevInfoTypes type = static_cast<DevInfoTypes>(*it);
+    std::cout << (std::to_string(*it) + " (" + devInfoTypesStrings.at(type)
+                  + ")");
+    auto temp_it = it;
+    if(++temp_it != env_vars_.enum_overrides.end()) {
+      std::cout << ", ";
+    }
+  }
+  std::cout << "}" << std::endl;
+}
+
 std::shared_ptr<Monitor>
 RocmSMI::FindMonitor(std::string monitor_path) {
   std::string tmp;
@@ -470,15 +654,10 @@ static const uint32_t kAmdGpuId = 0x1002;
 
 static bool isAMDGPU(std::string dev_path) {
   std::string vend_path = dev_path + "/device/vendor";
-  std::string vbios_v_path = dev_path + "/device/vbios_version";
   if (!FileExists(vend_path.c_str())) {
     return false;
   }
 
-  if (!FileExists(vbios_v_path.c_str())) {
-    return false;
-  }
-
   std::ifstream fs;
   fs.open(vend_path);
 
